/**
 * Copyright (C) 2011 Ovea <dev@ovea.com>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.ovea.resource.web;

import com.ovea.markup.CachingTemplateCompiler;
import com.ovea.markup.Template;
import com.ovea.markup.TemplateCompiler;
import com.ovea.markup.mvel.MVEL2TemplateCompiler;
import com.ovea.resource.web.util.Gzip;
import com.ovea.resource.web.util.MimeTypes;
import com.ovea.resource.web.util.Resource;
import com.yahoo.platform.yui.compressor.CssCompressor;
import com.yahoo.platform.yui.compressor.JavaScriptCompressor;
import org.apache.commons.codec.digest.DigestUtils;
import org.mozilla.javascript.ErrorReporter;
import org.mozilla.javascript.EvaluatorException;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.StringReader;
import java.io.StringWriter;
import java.util.*;
import java.util.concurrent.Executor;
import java.util.concurrent.atomic.AtomicLong;
import java.util.logging.Level;
import java.util.logging.Logger;

final class VirtualResource {

    private static Logger LOGGER = Logger.getLogger(VirtualResource.class.getName());
    private static final List<String> MARKUP_CONTENT_TYPES = Arrays.asList(
        "text/html",
        "text/css",
        "application/xml");

    private final AtomicLong lastModificationTimeInSeconds = new AtomicLong(0);
    private final List<WebResource> webResources = new LinkedList<WebResource>();
    private final String virtualPath;
    private final boolean debug;
    private final TemplateCompiler compiler;

    private String contentType = MimeTypes.UNKNOWN;
    private long caching = -1;
    private boolean gzip;
    private byte[] content;
    private byte[] gzipContent;

    private String etag;
    private Properties properties = new Properties();
    private Executor executor = new Executor() {
        @Override
        public void execute(Runnable command) {
            command.run();
        }
    };


    VirtualResource(String virtualPath, boolean debug) {
        this.virtualPath = virtualPath;
        this.debug = debug;
        this.compiler = debug ? new MVEL2TemplateCompiler() : new CachingTemplateCompiler(new MVEL2TemplateCompiler());
        setContentType(MimeTypes.getContentTypeForPath(virtualPath));
    }

    public void setProperties(Properties properties) {
        this.properties = properties;
    }

    void setContentType(String contentType) {
        this.contentType = contentType;
    }

    public boolean isMarkup() {
        return MARKUP_CONTENT_TYPES.contains(getContentType());
    }

    @Override
    public String toString() {
        return virtualPath;
    }

    public String getVirtualPath() {
        return virtualPath;
    }

    void setCaching(long caching) {
        this.caching = Math.max(caching, 0);
    }

    void setGzip(boolean gzip) {
        this.gzip = gzip;
    }

    public WebResource addWebResource(Resource resource) {
        if (LOGGER.isLoggable(Level.FINE))
            LOGGER.fine(virtualPath + " : + " + resource);
        WebResource webResource = new WebResource(resource);
        webResources.add(webResource);
        return webResource;
    }

    public boolean needsCaching() {
        return caching > 0;
    }

    /**
     * last modif time in milliseconds, rouned to the second
     */
    public long getLastModificationTime() {
        return lastModificationTimeInSeconds.get();
    }

    /**
     * caching time in seconds
     */
    public long getCaching() {
        return caching;
    }

    public String getContentType() {
        return contentType;
    }

    public boolean wantsGzip() {
        return gzip;
    }

    public byte[] getContent() {
        return content != null ? content : buildContent();
    }

    public byte[] getGzipContent() {
        return gzipContent != null ? gzipContent : buildGzipContent();
    }

    public void preload() {
        content = buildContent();
        if (gzip)
            gzipContent = buildGzipContent();
    }

    private byte[] buildContent() {
        if (LOGGER.isLoggable(Level.INFO))
            LOGGER.info("[" + virtualPath + "] Loading...");
        ByteArrayOutputStream baos = new ByteArrayOutputStream();
        // aggregate resources
        for (WebResource webResource : webResources) {
            if (LOGGER.isLoggable(Level.FINE))
                LOGGER.fine("[" + virtualPath + "] Loading " + webResource + "...");
            try {
                byte[] data;
                if (isMarkup()) {
                    Template template = compiler.compile(convert(webResource.location()));
                    data = template.merge(new HashMap<Object, Object>() {{
                        put("props", properties);
                    }}).getBytes();
                } else {
                    data = webResource.getData();
                }
                if (!debug && webResource.mustMinimize()) {
                    data = minimize(
                        webResource,
                        new String(data) + ("application/x-javascript".equals(contentType) || "text/css".equals(contentType) ? "\r\n" : "")).getBytes();
                }
                baos.write(data);
            } catch (Exception e) {
                throw new RuntimeException("Error building content of virtual resource " + virtualPath + " when processing resource " + webResource + " : " + e.getMessage(), e);
            }
        }
        modified();
        return baos.toByteArray();
    }

    private byte[] buildGzipContent() {
        return Gzip.compress(getContent());
    }

    private String minimize(final WebResource webResource, String data) {
        if ("text/css".equals(contentType)) {
            if (LOGGER.isLoggable(Level.FINE))
                LOGGER.fine("[" + virtualPath + "] Minifying CSS " + webResource + "...");
            try {
                CssCompressor compressor = new CssCompressor(new StringReader(data));
                StringWriter sw = new StringWriter();
                compressor.compress(sw, -1);
                return sw.toString();
            } catch (IOException e) {
                throw new IllegalStateException("[" + virtualPath + "] Unable to minimize CSS " + webResource + " : " + e.getMessage(), e);
            }
        } else if ("application/x-javascript".equals(contentType)) {
            if (LOGGER.isLoggable(Level.FINE))
                LOGGER.fine("[" + virtualPath + "] Minifying Javascript...");
            try {
                JavaScriptCompressor compressor = new JavaScriptCompressor(new StringReader(data), new ErrorReporter() {
                    @Override
                    public void warning(String message, String sourceName, int line, String lineSource, int lineOffset) {
                        if (line < 0) {
                            if (LOGGER.isLoggable(Level.FINE))
                                LOGGER.fine("[" + virtualPath + "] Minifying Javascript " + webResource + " : " + message);
                        } else {
                            if (LOGGER.isLoggable(Level.FINE))
                                LOGGER.fine("[" + virtualPath + "] Minifying Javascript " + webResource + " : line " + line + ", offset " + lineOffset + " : " + message);
                        }
                    }

                    @Override
                    public void error(String message, String sourceName, int line, String lineSource, int lineOffset) {
                        if (line < 0) {
                            LOGGER.severe("[" + virtualPath + "] Minimizing Javascript " + webResource + " : " + message);
                        } else {
                            LOGGER.severe("[" + virtualPath + "] Minimizing Javascript " + webResource + " : " + line + ':' + lineOffset + ':' + message);
                        }
                    }

                    @Override
                    public EvaluatorException runtimeError(String message, String sourceName, int line, String lineSource, int lineOffset) {
                        error(message, sourceName, line, lineSource, lineOffset);
                        return new EvaluatorException(message);
                    }
                });
                StringWriter sw = new StringWriter();
                compressor.compress(sw, -1, false, true, false, false);
                return sw.toString();
            } catch (IOException e) {
                throw new IllegalStateException("[" + virtualPath + "] Unable to minimize Javascript " + webResource + " : " + e.getMessage(), e);
            }
        } else {
            return data;
        }
    }

    private com.ovea.markup.util.Resource convert(Resource resource) {
        return resource.isURL() ?
            com.ovea.markup.util.Resource.from(resource.toURL()) :
            com.ovea.markup.util.Resource.fromData(resource.readAsBytes());
    }

    private void modified() {
        long time = System.currentTimeMillis();
        time = time - (time % 1000);
        lastModificationTimeInSeconds.set(time);
    }

    public String getETag() {
        if (etag == null) {
            etag = "\"" + DigestUtils.md5Hex(getContent()) + "\"";
        }
        return etag;
    }

    public void setExecutor(Executor executor) {
        this.executor = executor;
    }
}