/**
 * Copyright (C) 2011 Ovea <dev@ovea.com>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.ovea.resource.web;

import java.util.LinkedHashMap;
import java.util.Map;
import java.util.Properties;
import java.util.concurrent.*;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.logging.Logger;

final class VirtualResources {

    private static Logger LOGGER = Logger.getLogger(VirtualResources.class.getName());

    private final Map<String, VirtualResource> virtualResources = new LinkedHashMap<String, VirtualResource>();
    private boolean debug;
    private Properties properties = new Properties();
    private final ThreadPoolExecutor executor = new ThreadPoolExecutor(
        10, Integer.MAX_VALUE,
        60L, TimeUnit.SECONDS,
        new SynchronousQueue<Runnable>(),
        new ThreadFactory() {
            private final ThreadGroup group = Thread.currentThread().getThreadGroup();
            private final AtomicInteger threadNumber = new AtomicInteger(1);
            private final String namePrefix = "VirtualResourceLoader-";

            @Override
            public Thread newThread(Runnable r) {
                Thread t = new Thread(group, r, namePrefix + threadNumber.getAndIncrement(), 0);
                if (t.isDaemon())
                    t.setDaemon(false);
                if (t.getPriority() != Thread.NORM_PRIORITY)
                    t.setPriority(Thread.NORM_PRIORITY);
                return t;
            }
        }, new ThreadPoolExecutor.CallerRunsPolicy());

    VirtualResource addVirtualResource(String virtualPath) {
        if (!virtualPath.startsWith("/"))
            virtualPath = "/" + virtualPath;
        virtualPath = virtualPath.toLowerCase();
        if (virtualResources.containsKey(virtualPath))
            throw new IllegalStateException("Virtual path " + virtualPath + " already exist");
        LOGGER.fine(virtualPath);
        VirtualResource virtualResource = new VirtualResource(virtualPath, debug);
        virtualResource.setProperties(properties);
        virtualResource.setExecutor(executor);
        virtualResources.put(virtualPath, virtualResource);
        return virtualResource;
    }

    public void setProperties(Properties properties) {
        this.properties = properties;
    }

    public void setDebug(boolean debug) {
        this.debug = debug;
    }

    @Override
    public String toString() {
        return virtualResources.values().toString();
    }

    public void preload() {
        CompletionService<Object> completionService = new ExecutorCompletionService<Object>(executor);
        long count = 0;
        for (final VirtualResource virtualResource : virtualResources.values()) {
            completionService.submit(new Runnable() {
                @Override
                public void run() {
                    virtualResource.preload();
                }
            }, null);
            count++;
        }
        while (count-- > 0) {
            try {
                completionService.take().get();
            } catch (InterruptedException e) {
                Thread.currentThread().interrupt();
                return;
            } catch (ExecutionException e) {
                Throwable t = e.getCause();
                if(t instanceof RuntimeException) throw (RuntimeException) t;
                if(t instanceof Error) throw (Error) t;
                throw new RuntimeException(e.getMessage(), e);
            }
        }
    }

    public boolean has(String path) {
        return virtualResources.containsKey(path.toLowerCase());
    }

    public VirtualResource get(String path) {
        VirtualResource virtualResource = virtualResources.get(path.toLowerCase());
        if (virtualResource == null)
            throw new IllegalStateException("Unknown virtual resource: " + path);
        return virtualResource;
    }
}